/**
 * Binary Search Tree with keys of type E (which are Comparable).
 */
public class BinarySearchTree<E extends Comparable<E>> {

  /**
   * Internal class to represent a node in a binary search tree.
   */
  class Node {
    public E key;
    public Node left;
    public Node right;
    public int height;

    /**
     * Constructs a BST node, saving the @param key.
     */
    public Node(E key) {
      this.key = key;
    }

    public void calculateHeight() {
      // DONE
      // height of null node is -1
      // height of leaf node is 0
      // otherwise height is 1 + max(leftHeight, rightHeight)
      int hL = (left == null) ? -1 : left.height;
      int hR = (right == null) ? -1 : right.height;
      height = 1 + Math.max(hL, hR);
    }
  }

  // root of the tree
  private Node root;

  /**
   * Constructs an empty binary tree.
   */
  public BinarySearchTree() {
    root = null;
  }

  /**
   * Determines if this BST contains a particular key.
   * 
   * @param key - key to look for.
   * @return true/false if one of the nodes in this tree has this key.
   */
  public boolean contains(E key) {
    return contains(root, key);
  }

  private boolean contains(Node node, E key) {
    if (node == null) return false;
    int compareResult = key.compareTo(node.key);
    if (compareResult == 0) return true;
    else if (compareResult < 0) return contains(node.left, key);
    else return contains(node.right, key);
  }

  /**
   * Adds a key to the BST if not already at a node in the tree.
   * 
   * @param key - key to save at a newly created node.
   */
  public void add(E key) {
    root = add(root, key);
  }

  private Node add(Node node, E key) {
    if (node == null) return new Node(key);
    int compareResult = key.compareTo(node.key);
    if (compareResult < 0) {
      node.left = add(node.left, key);
    } else if (compareResult > 0) {
      node.right = add(node.right, key);
    }
    node.calculateHeight();
    return node;
  }

  /**
   * Removes a key from the BST (if it exists).
   * 
   * @param key - key to remove from the tree.
   */
  public void remove(E key) {
    root = remove(root, key);
  }

  private Node remove(Node node, E key) {
    if (node == null) {
      return null; // went beyond a leaf, key not found
    }
    int compareResult = key.compareTo(node.key);
    if (compareResult < 0) {
      // node to remove is (possibly) in the left subtree
      node.left = remove(node.left, key);
    } else if (compareResult > 0) {
      // node to remove is (possibly) in the right subtree
      node.right = remove(node.right, key);
    } else {
      // we are at the node to delete, check 3 cases
      if (node.left == null && node.right == null) {
        return null; // case 1, delete a leaf
      } else if (node.left == null) {
        return node.right; // case 2a
      } else if (node.right == null) {
        return node.left; // case 2b
      } else {
        // case 3: two children
        Node successor = getMinNode(node.right);
        node.key = successor.key;
        node.right = remove(node.right, successor.key);
      }
    }
    node.calculateHeight();
    return node;
  }

  /**
   * Finds the minimum node in a tree rooted at a particular node.
   * 
   * @param node - root of the (sub)tree to find the minimum of.
   * @return node with the minimum key.
   */
  public Node getMinNode(Node node) {
    while (node.left != null) {
      node = node.left;
    }
    return node;
  }

  /**
   * @return a String representation of the tree.
   */
  public String toString() {
    return toString("", root);
  }

  /**
   * Returns a string representation of the subtree.
   * 
   * @param padding - spacing and vertical bars to print before the node.
   * @param node    - the root of the subtree to print.
   * @return appended line to print for this node (and child nodes).
   */
  private String toString(String padding, Node node) {
    if (node == null) {
      return "";
    }

    String result = padding + "└──(" + node.key.toString() + ")" + "\n";

    padding += "│    ";
    result += toString(padding, node.left);
    result += toString(padding, node.right);

    return result;
  }

  /**
   * Test some BST methods.
   * 
   * @param args
   */
  public static void main(String[] args) {
    BinarySearchTree<Integer> bst = new BinarySearchTree<>();

    bst.add(12);
    bst.add(8);
    bst.add(14);
    bst.add(5);
    bst.add(9);
    bst.add(20);
    bst.add(17);

    System.out.println(bst);

    bst.remove(12);
    System.out.println(bst);
  }
}